/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <mach/machine/asm.h>
#include <mach/machine/eflags.h>

#include "asm.h"
#include "trap.h"
#include "config.h"
#include "i16_crt0.h"
#include "debug.h"

#ifdef ENABLE_DPMI

#define EXCHANDLER(n)				\
ENTRY(dpmi_exchandler_##n)			;\
	pushl	$n				;\
	jmp	dpmi_exchandler_common

EXCHANDLER(0)
EXCHANDLER(1)
EXCHANDLER(2)
EXCHANDLER(3)
EXCHANDLER(4)
EXCHANDLER(5)
EXCHANDLER(6)
EXCHANDLER(7)
EXCHANDLER(8)
EXCHANDLER(9)
EXCHANDLER(10)
EXCHANDLER(11)
EXCHANDLER(12)
EXCHANDLER(13)
EXCHANDLER(14)
EXCHANDLER(15)

ENTRY(dpmi_exchandler_common)
	pushl	%eax
	pushl	%ebx
	pushl	%ds

	movl	%cs:EXT(dpmi_kernel_stack)+4,%eax
	movw	%ax,%ds

	/* Switch to the kernel stack if we aren't on it already.  */
	movl	10*4(%esp),%ebx
	cmpw	%ax,11*4(%esp)
	je	1f
	movl	EXT(dpmi_kernel_stack),%ebx
1:

	/* Push the bottom part of the trap state frame on the kernel stack.  */
	subl	$7*4,%ebx
	movl	3*4(%esp),%eax
	movl	%eax,(%ebx)
	movl	6*4(%esp),%eax
	movl	%eax,1*4(%ebx)
	movl	7*4(%esp),%eax
	movl	%eax,2*4(%ebx)
	movl	8*4(%esp),%eax
	movl	%eax,3*4(%ebx)
	movl	9*4(%esp),%eax
	movl	%eax,4*4(%ebx)
	movl	10*4(%esp),%eax
	movl	%eax,5*4(%ebx)
	movl	11*4(%esp),%eax
	movl	%eax,6*4(%ebx)

	/* Adjust the return information to point to
	   the second half of our exception handler.  */
	movl	$1f,7*4(%esp)
	movw	%cs,%ax
	movl	%eax,8*4(%esp)
	andl	$-1-(EFL_IF | EFL_TF | EFL_DF),9*4(%esp)
	movl	%ebx,10*4(%esp)
	movw	%ds,%ax
	movl	%eax,11*4(%esp)

	/* Return to the DPMI host.  */
	popl	%ds
	popl	%ebx
	popl	%eax
	addl	$4,%esp
	lret

1:
	pusha
	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* If we trapped in the kernel, just die.  */
	movw	%cs,%ax
	cmpw	%ax,TR_CS(%esp)
	jne	1f
	pushl	%esp
	call	EXT(trap_dump_die);
1:

	/* Convert the trap into an appropriate signal.  */
	cmpl	$0x20,TR_TRAPNO(%esp)
	jae	1f
	call	EXT(sig_trap)
1:

	/* Check for new signals before returning to the application.  */
	movl	EXT(new_sigs),%edx
	orl	%edx,%edx
	jz	1f
dpmi_check_sigs_and_return:
	call	EXT(check_sigs)
1:

	/* Restore saved state and return.  */
A(e,	cmpl	$EXT(crt0_stack)+CRT0_STACK_SIZE-TR_USIZE,%esp)
	popl	%gs
	popl	%fs
	popl	%es
	popl	%ds
	popa
	addl	$4*4,%esp
	testl	$EFL_IF,(%esp)
	jnz	1f
	popfl
	lss	%cs:(EXT(crt0_stack)+CRT0_STACK_SIZE-TR_USIZE+TR_ESP),%esp
	ljmp	%cs:(EXT(crt0_stack)+CRT0_STACK_SIZE-TR_USIZE+TR_EIP)
1:
	andl	$-1-EFL_IF,(%esp)
	popfl
	lss	%cs:(EXT(crt0_stack)+CRT0_STACK_SIZE-TR_USIZE+TR_ESP),%esp
	sti
	ljmp	%cs:(EXT(crt0_stack)+CRT0_STACK_SIZE-TR_USIZE+TR_EIP)


/*
 * Hardware interrupt handlers for reflecting interrupts to the protected mode program.
 */
#define DELAY							\
	jmp	9f						;\
9:	jmp	9f						;\
9:

#define ACK_FIRST(n)						\
	movb	$0x20,%al					;\
	outb	%al,$0x20

#define ACK_SECOND(n)						\
	movb	$0x20,%al					;\
	outb	%al,$0xa0					;\
	DELAY							;\
	outb	%al,$0x20

/* It appears DPMI is just plain too stupid
   to support dispatching interrupts as signals,
   because it doesn't define anything about the interrupt stack frame
   that allows us to redirect the flow of the program if it's in user mode.  */
#if 0
#define BLK_FIRST(n)						\
	inb	$0x21,%al					;\
	DELAY							;\
	orb	$1<<n,%al					;\
	outb	%al,$0x21					;\

#define BLK_SECOND(n)						\
	inb	$0xa1,%al					;\
	DELAY							;\
	orb	$1<<(n-8),%al					;\
	outb	%al,$0xa1					;\

#define UNBLK_FIRST(n)						\
	inb	$0x21,%al					;\
	DELAY							;\
	andb	$-1-(1<<n),%al					;\
	outb	%al,$0x21

#define UNBLK_SECOND(n)						\
	inb	$0xa1,%al					;\
	DELAY							;\
	andb	$-1-(1<<(n-8)),%al				;\
	outb	%al,$0xa1

#define	SIG_INTHANDLER(n,pic)					\
	.data							;\
	.long	1f						;\
	.data	2						;\
	.long	2f						;\
	.text							;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* dpmi_sig_int_handler */				;\
1:	pushl	$0x20+n						;\
	pushl	%eax						;\
	pushl	%ebx						;\
	leal	EXT(sig_irq_nodes)+(SN_SIZE*n),%ebx		;\
	BLK_##pic(n)						;\
	DELAY							;\
	ACK_##pic(n)						;\
	jmp	dpmi_sig_intrs					;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* dpmi_sig_int_post_handler */				;\
2:	UNBLK_##pic(n)						;\
	ret

	.data
DATA(dpmi_sig_int_handlers)
	.data	2
DATA(dpmi_sig_int_post_handlers)

SIG_INTHANDLER(0,FIRST)
SIG_INTHANDLER(1,FIRST)
SIG_INTHANDLER(2,FIRST)
SIG_INTHANDLER(3,FIRST)
SIG_INTHANDLER(4,FIRST)
SIG_INTHANDLER(5,FIRST)
SIG_INTHANDLER(6,FIRST)
SIG_INTHANDLER(7,FIRST)
SIG_INTHANDLER(8,SECOND)
SIG_INTHANDLER(9,SECOND)
SIG_INTHANDLER(10,SECOND)
SIG_INTHANDLER(11,SECOND)
SIG_INTHANDLER(12,SECOND)
SIG_INTHANDLER(13,SECOND)
SIG_INTHANDLER(14,SECOND)
SIG_INTHANDLER(15,SECOND)

	.text
	P2ALIGN(TEXT_ALIGN)
dpmi_sig_intrs:
	pushl	%ds

	/* Find the kernel's data segment (ds).  */
	movl	%cs:EXT(dpmi_kernel_stack)+4,%eax
	movw	%ax,%ds

	/* Add the appropriate signode (ebx) to the pending queue.  */
	pushl	%ecx
	lea	EXT(sig_pending_list),%eax
	movl	SN_NEXT(%eax),%ecx
	movl	%ecx,SN_NEXT(%ebx)
	movl	%eax,SN_PREV(%ebx)
	movl	%ebx,SN_NEXT(%eax)
	movl	%ebx,SN_PREV(%ecx)
	popl	%ecx

	/* Indicate that new signals need to be checked.  */
	movl	%esp,EXT(new_sigs)

	/* If we were running code, then take the signal right away.
	   Otherwise, defer it until we return back into user mode.
	   (Check the code segment rather than the stack
	   because that's the first thing changed when coming in on a syscall
	   and the last thing to be changed back when going back out...)  */
	movw	EXT(dpmi_prog_cs),%ax
	cmpw	%ax,8*4(%esp)
	jne	1f

	/* Even though the DPMI (0.9) spec refers to interrupt handlers
	   as "exception handlers", Windows doesn't give us a full
	   exception handler stack frame.  Instead we get only:
	   	cs
		eip
		eflags
		ret cs
		ret eip
	   Linux DOSEMU seems to follow this convention as well,
	   although it keeps a separate copy of eflags stashed elsewhere
	   so changes to the stacked eflags don't matter.
	   Hopefully other OS's work similarly as well.  */
	movl	7*4(%esp),%eax
	movl	%eax,EXT(dpmi_int_eip_save)
	movl	$dpmi_int_tail,7*4(%esp)
	movw	%cs,8*4(%esp)

1:
	popl	%ds
	popl	%ebx
	popl	%eax
	addl	$4,%esp		/* pop the irq number */
	lret

dpmi_int_tail:
ljmp $0,$0
	pushfl
	cli
	pushl	%eax
	pushl	%ebx

	/* Switch to the kernel stack,
	   saving the user stack in ebx.  */
	movl	%esp,%ebx
	lss	%cs:(EXT(dpmi_kernel_stack)),%esp
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es
	movw	EXT(dpmi_prog_stack+4),%fs

	/* Build the trap frame.  */
	pushl	EXT(dpmi_prog_ds)
	addl	$12,%ebx
	pushl	%ebx
	pushl	%fs:-4(%ebx)
	pushl	EXT(dpmi_prog_cs)
	pushl	EXT(dpmi_int_eip_save)
	subl	$2*4,%esp
	pushl	%fs:-8(%ebx)
	pushl	%ecx
	pushl	%edx
	pushl	%fs:-12(%ebx)
	subl	$4,%esp
	pushl	%ebp
	pushl	%esi
	pushl	%edi
	pushl	%fs
	pushl	%fs
	pushl	$0
	pushl	$0
	orl	$EFL_IF,TR_EFLAGS(%esp)

	pushl	%esp
	call	EXT(trap_dump_die);

	.comm	dpmi_int_eip_save,4
#endif 0

#define FASTINTHANDLER(n,pic)					\
	.data							;\
	.long	1f						;\
	.text							;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* dpmi_fastint_handler */				;\
1:	pushl	%eax						;\
	ACK_##pic(n)						;\
	leal	EXT(dpmi_fastint_nodes)+(8*n),%eax		;\
	jmp	dpmi_fastint_common

	.data
DATA(dpmi_fastint_handlers)
FASTINTHANDLER(0,FIRST)
FASTINTHANDLER(1,FIRST)
FASTINTHANDLER(2,FIRST)
FASTINTHANDLER(3,FIRST)
FASTINTHANDLER(4,FIRST)
FASTINTHANDLER(5,FIRST)
FASTINTHANDLER(6,FIRST)
FASTINTHANDLER(7,FIRST)
FASTINTHANDLER(8,SECOND)
FASTINTHANDLER(9,SECOND)
FASTINTHANDLER(10,SECOND)
FASTINTHANDLER(11,SECOND)
FASTINTHANDLER(12,SECOND)
FASTINTHANDLER(13,SECOND)
FASTINTHANDLER(14,SECOND)
FASTINTHANDLER(15,SECOND)

	.text
	P2ALIGN(TEXT_ALIGN)
dpmi_fastint_common:
	pushl	%ecx
	pushl	%edx
	pushl	%ds
	pushl	%es

	/* Load the user's data segment registers.  */
	movw	%cs:EXT(dpmi_prog_ds),%dx
	movw	%dx,%ds
	movw	%dx,%es

	/* Find the user's fastint handler parameters.
	   The handler ESP has already been adjusted downward
	   to make room for the stuff we're going to push onto it.  */
	movl	%cs:(%eax),%ecx		/* handler EIP */
	movl	%cs:4(%eax),%eax	/* handler ESP */

	/* Set up the long return stack frame for entry into the handler.  */
	movl	%ecx,(%eax)
	movw	%cs:EXT(dpmi_prog_cs),%cx
	movl	%ecx,4(%eax)

	/* Stash our current stack pointer so we'll be able to get back.  */
	movl	%esp,12(%eax)
	movw	%ss,16(%eax)

	/* Switch to the handler stack.  */
	movl	%eax,%esp
	movw	%dx,%ss

	/* Drop into the handler.  */
	cld
	lret

ENTRY(dpmi_fastint_return)

	/* Get back to the kernel stack.  */
	lss	(%esp),%esp

	/* Restore the saved state and return from the interrupt.  */
	popl	%es
	popl	%ds
	popl	%edx
	popl	%ecx
	popl	%eax
	sti
	iret


ENTRY(dpmi_syscall_entry)
	cli
	cld

	/* Check for the special sigreturn syscall.  */
	cmpl	$-1,%eax
	jz	dpmi_sigreturn

	movl	%esp,%ecx
	lss	%cs:(EXT(dpmi_kernel_stack)),%esp
	movw	%ss,%dx
	movw	%dx,%ds
	movw	%dx,%es
	movl	%ecx,EXT(dpmi_prog_stack)

	movl	$0,EXT(errno)

	addl	EXT(prog_va),%ecx
	addl	$12,%ecx
	pushl	%ecx

	call	*EXT(syscall_table)(,%eax,4)

	movl	EXT(errno),%edx

	/* Check for new signals that arrived during the syscall.  */
	movl	EXT(new_sigs),%ecx
	orl	%ecx,%ecx
	jnz	1f

	lss	(EXT(dpmi_prog_stack)),%esp
	movw	%ss,%cx
	movw	%cx,%ds
	movw	%cx,%es

	sti
	lret

	/* New signals have arrived.  Build a trap frame.  */
1:
	movl	$EXT(crt0_stack)+CRT0_STACK_SIZE-TR_USIZE+TR_TRAPNO,%esp
	lfs	(EXT(dpmi_prog_stack)),%ecx
	pusha
	pushl	%fs
	pushl	%fs
	pushl	$0
	pushl	$0
	movl	%fs:(%ecx),%eax		/* eip */
	movl	%eax,TR_EIP(%esp)
	movl	%fs:4(%ecx),%eax	/* cs */
	movl	%eax,TR_CS(%esp)
	pushfl				/* eflags */
	popl	%eax
	orl	$EFL_IF,%eax
	movl	%eax,TR_EFLAGS(%esp)
	addl	$8,%ecx			/* esp */
	movl	%ecx,TR_ESP(%esp)
	movw	%fs,TR_SS(%esp)		/* ss */

	jmp	dpmi_check_sigs_and_return

dpmi_sigreturn:
	ljmp	$0,$0

#endif ENABLE_DPMI
