/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <mach/boolean.h>
#include <mach/machine/code16.h>
#include <mach/machine/vm_param.h>
#include <mach/machine/proc_reg.h>
#include <mach/machine/eflags.h>
#include <mach/machine/asm.h>

#include "trap.h"
#include "gdt.h"
#include "cpu.h"
#include "real_tss.h"
#include "debug.h"


/* Small dummy "kernel" stack to use while in v86 mode.
   v86_kstack[1] is the actual trap frame;
   v86_kstack[0] is some extra stack space
   for handling unexpected traps from v86 mode.  */
static struct trap_state v86_kstack[2];


/* Nonzero if we're in v86 mode.  */
static boolean_t in_v86_mode;


CODE16

void i16_v86_switch_to_real_mode(int foo)
{
	extern unsigned short real_cs;

	/* Make sure interrupts are disabled.  */
	i16_cli();

	/* Make sure the real_tss points to the dummy v86_kstack.  */
	real_tss.esp0 = (vm_offset_t)(&v86_kstack[2]);

	/* Switch to the LDT to use in v86 mode.  */
	cpu[0].tables.gdt[REAL_TSS_IDX].access &= ~ACC_TSS_BUSY;
	i16_set_tr(REAL_TSS);

	/* Drop into v86 mode.  */
	asm volatile("
		movl	%%esp,%%eax
		pushl	$0
		pushl	$0
		pushl	%0
		pushl	%0
		pushl	%0
		pushl	%%eax
		pushl	%1
		pushl	%0
		pushl	$1f
		iret
	1:
	" :
	  : "b" (real_cs), "c" (get_eflags() | EFL_VM)
	  : "eax");
}

asm("
	.globl	"SEXT(i16_v86_switch_to_pmode)"
"SEXT(i16_v86_switch_to_pmode)":
	movb	$0x89,%ah
	int	$0x15
");

void i16_v86_switch_to_pmode_finish()
{
	/* Switch to the normal kernel TSS.  */
	cpu[0].tables.gdt[KERNEL_TSS_IDX].access &= ~ACC_TSS_BUSY;
	i16_set_tr(KERNEL_TSS);
}

