/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1993,1992,1991,1990 Carnegie Mellon University
 * Copyright (c) 1991 IBM Corporation 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation,
 * and that the nema IBM not be used in advertising or publicity 
 * pertaining to distribution of the software without specific, written
 * prior permission.
 * 
 * CARNEGIE MELLON AND IBM ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND IBM DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */

#include <mach/machine/asm.h>
#include <mach/machine/eflags.h>
#include <mach/machine/trap.h>
#include <mach/machine/seg.h>

#include "asm.h"
#include "trap.h"
#include "debug.h"
#include "config.h"


/*
 * Trap/interrupt entry points.
 *
 * All traps must create the following save area on the kernel stack:
 *
 *	gs
 *	fs
 *	es
 *	ds
 *	edi
 *	esi
 *	ebp
 *	cr2 if page fault - otherwise unused
 *	ebx
 *	edx
 *	ecx
 *	eax
 *	trap number
 *	error code
 *	eip
 *	cs
 *	eflags
 *	user esp - if from user
 *	user ss  - if from user
 *	es       - if from V86 thread
 *	ds       - if from V86 thread
 *	fs       - if from V86 thread
 *	gs       - if from V86 thread
 *
 */

/*
 * Page fault traps save cr2.
 */
ENTRY(t_page_fault)
	pushl	$(T_PAGE_FAULT)		/* mark a page fault trap */
	pusha				/* save the general registers */

	movl	%cr2,%eax		/* get the faulting address */
	movl	%eax,12(%esp)		/* save in esp save slot */

	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	/* Please no page faults in v86 mode... */
A(z,	testl	$EFL_VM,TR_EFLAGS(%esp))

	/* Load the kernel segment registers.  */
	movw	%ss,%dx
	movw	%dx,%ds
	movw	%dx,%es

	/* Try to handle the page fault.  */
	pushl	$0	/* not locking */
	pushl	%eax
	call	EXT(page_fault)
	addl	$8,%esp
	orl	%eax,%eax
	jz	return_from_trap

	/* Couldn't be handled - see if there's a recovery routine.  */
	.comm	recovery,4
	movl	recovery,%eax
	orl	%eax,%eax
	jz	1f
	movl	%eax,TR_EIP(%esp)
	jmp	return_from_trap

1:
	UNEXPECTED_TRAP

/*
 * General protection fault reflects 'int' instructions.
 */
ENTRY(t_gen_prot)
	pushl	$(T_GENERAL_PROTECTION)
	pusha
	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	/* See if we came from v86 mode.  */
	testl	$EFL_VM,TR_EFLAGS(%esp)
	jnz	gpf_from_v86

/* Trapped from protected mode.  */
gpf_from_pmode:

	/* The kernel should never produce GPFs.  */
	testb	$3,TR_CS(%esp)
	jz	1f

	/* Check the faulting instruction (al).
	   Assume ds is the same as faulting cs.  */
	movl	TR_EIP(%esp),%edx
	movb	(%edx),%al
	cmpb	$0xcd,%al
	je	gpf_int_n_pmode

1:
	UNEXPECTED_TRAP

gpf_int_n_pmode:

	/* Bump the ip past the int instruction.  */
	addl	$2,TR_EIP(%esp)

	/* Find the real mode interrupt vector number,
	   and patch the int instruction below.  */
	movb	1(%edx),%al
	movb	%al,%ss:2f+1

	/* Load the kernel segment registers.  */
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* Switch to v86 mode and handle the interrupt.  */
	ljmp	$KERNEL_16_CS,$1f
	.code16
1:	call	*EXT(i16_switch_to_real_mode)

	movl	TR_EAX(%esp),%eax
	movl	TR_EBX(%esp),%ebx
	movl	TR_ECX(%esp),%ecx
	movl	TR_EDX(%esp),%edx
	movl	TR_ESI(%esp),%esi
	movl	TR_EDI(%esp),%edi
	movl	TR_EBP(%esp),%ebp

2:	int	$0

	movl	%eax,TR_EAX(%esp)
	movl	%ebx,TR_EBX(%esp)
	movl	%ecx,TR_ECX(%esp)
	movl	%edx,TR_EDX(%esp)
	movl	%esi,TR_ESI(%esp)
	movl	%edi,TR_EDI(%esp)
	movl	%ebp,TR_EBP(%esp)

	pushf
	popl	%eax
	movl	TR_EFLAGS(%esp),%edx
	andl	$EFL_IF,%edx
	orl	%edx,%eax
	movw	%ax,TR_EFLAGS(%esp)

	call	*EXT(i16_switch_to_pmode)
	ljmp	$KERNEL_CS,$return_from_trap
	.code32


/* Trapped from v86 mode.  */
gpf_from_v86:

	/* Load the linear/physical-address data segment,
	   for easy access to real-mode memory.  */
	movl	$LINEAR_DS,%eax
	movw	%ax,%ds

	/* Find the physical address of the trapping instruction (ebx).  */
	movzwl	TR_CS(%esp),%ebx
	shll	$4,%ebx
	addl	TR_EIP(%esp),%ebx

	/* Check the instruction (al).  */
	movb	(%ebx),%al
	cmpb	$0xcd,%al
	je	gpf_int_n

	UNEXPECTED_TRAP


gpf_int_n:

	/* Bump the ip past the int instruction.  */
	addw	$2,TR_EIP(%esp)

	/* Find the real mode interrupt vector number (esi).  */
	movzbl	1(%ebx),%esi

	/* See if it's a copy-extended-memory interrupt request;
	   if so, just handle it here.  */
	cmpl	$0x15,%esi
	jne	1f
	cmpb	$0x87,TR_EAX+1(%esp)
	je	copy_ext_mem
	cmpb	$0x89,TR_EAX+1(%esp)
	je	switch_from_v86
1:

reflect_v86_intr:
	/* Find the address of the real mode interrupt vector (esi).  */
	shll	$2,%esi

	/* Make room for the real-mode interrupt stack frame.  */
	subw	$6,TR_ESP(%esp)

	/* Find the physical address of the v86 stack (ebx).  */
	movzwl	TR_SS(%esp),%ebx
	shll	$4,%ebx
	addl	TR_ESP(%esp),%ebx

	/* Store the return information into the v86 stack frame.  */
	movl	TR_EIP(%esp),%eax
	movw	%ax,(%ebx)
	movl	TR_CS(%esp),%eax
	movw	%ax,2(%ebx)
	movl	TR_EFLAGS(%esp),%eax
	movw	%ax,4(%ebx)

	/* Find the real-mode interrupt vector to invoke,
	   and set up the real_call_thread's kernel stack frame
	   to point to it.  */
	movl	(%esi),%eax
	movw	%ax,TR_EIP(%esp)
	shrl	$16,%eax
	movw	%ax,TR_CS(%esp)
	andl	$-1-EFL_IF-EFL_TF,TR_EFLAGS(%esp)

	/* Restore saved state and return.  */
	addl	$4*4,%esp
	popa
	addl	$4*2,%esp
	iret



/* We're trying to switch out of v86 mode.
   Switch back to the normal stack
   and invoke i16_v86_switch_to_pmode_finish().
   Don't use any caller-saved registers.  */
switch_from_v86:

	/* Reload the callee-saved registers we trashed above.  */
	movl	TR_EBX(%esp),%ebx
	movl	TR_ESI(%esp),%esi

	/* Load the kernel segment registers.  */
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* Switch back to the normal kernel stack.  */
	movl	TR_ESP(%esp),%esp
	ljmp	$KERNEL_16_CS,$EXT(i16_v86_switch_to_pmode_finish)



/* We intercepted a copy-extended-memory software interrupt
   (int 0x15 function 0x87).
   The BIOS's routine isn't going to work in v86 mode,
   so do it ourselves.  */
copy_ext_mem:

	/* Find the parameter block provided by the caller (ebx).  */
	movzwl	TR_V86_ES(%esp),%ebx
	movzwl	TR_ESI(%esp),%eax
	shll	$4,%ebx
	addl	%eax,%ebx

	/* Source address (esi).  */
	movl	0x12(%ebx),%esi
	andl	$0x00ffffff,%esi

	/* Destination address (edi).  */
	movl	0x1a(%ebx),%edi
	andl	$0x00ffffff,%edi

	/* Number of bytes (ecx).  */
	movzwl	TR_ECX(%esp),%ecx
	addl	%ecx,%ecx

	/* Use the standard i386 bcopy routine to copy the data.
	   This assumes it's "friendly" in its use of segment registers
	   (i.e. always uses ss for stack data and ds/es for the data to copy).
	   The bcopy is simple enough that this should always be true.  */
	movw	%ds,%ax
	movw	%ax,%es
	cld
	pushl	%ecx
	pushl	%edi
	pushl	%esi
	call	EXT(bcopy)
	addl	$3*4,%esp

	/* Clear the carry flag to indicate that the copy was successful.
	   AH is also cleared, below.  */
	andl	$-1-EFL_CF,TR_EFLAGS(%esp)

	/* Restore saved state and return.  */
	addl	$4*4,%esp
	popa
	xorb	%ah,%ah
	addl	$4*2,%esp
	iret


/*
 * All 'exceptions' enter here with:
 *	esp->   trap number
 *		error code
 *		old eip
 *		old cs
 *		old eflags
 *		old esp		if trapped from user
 *		old ss		if trapped from user
 */
ENTRY(alltraps)
	pusha				/* save the general registers */
	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	/* See if the trap was from user-mode pmode code.
	   If not, it's a trap we can't handle - just die immediately.  */
	testl	$EFL_VM,TR_EFLAGS(%esp)
	jnz	1f
	testb	$3,TR_CS(%esp)
	jnz	1f
	UNEXPECTED_TRAP

1:
	/* Load the kernel segment registers.  */
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* Convert the trap to an appropriate signal.  */
	call	EXT(sig_trap)

return_from_trap:

	/* Check for new signals before returning to user mode.  */
	testl	$EFL_VM,TR_EFLAGS(%esp)
	jnz	1f
	testb	$3,TR_CS(%esp)
	jz	1f
	movl	EXT(new_sigs),%edx
	orl	%edx,%edx
	jnz	check_sigs_and_return
1:

	/* Restore saved state and return.  */
	popl	%gs
	popl	%fs
	popl	%es
	popl	%ds
	popa
	addl	$4*2,%esp
	iret


/*
 * Hardware interrupt handlers for reflecting interrupts to real/v86 mode.
 */
#define	REALINTHANDLER(n,real_n)					\
	.data							;\
	.long	1f						;\
	.text							;\
1:								;\
	pushl	$(real_n)					;\
	pushl	$0x20+n /* (just for debugging) */		;\
	jmp	allintrs

	.data
DATA(real_int_handlers)

REALINTHANDLER(0,0x08)
REALINTHANDLER(1,0x09)
REALINTHANDLER(2,0x0a)
REALINTHANDLER(3,0x0b)
REALINTHANDLER(4,0x0c)
REALINTHANDLER(5,0x0d)
REALINTHANDLER(6,0x0e)
REALINTHANDLER(7,0x0f)
REALINTHANDLER(8,0x70)
REALINTHANDLER(9,0x71)
REALINTHANDLER(10,0x72)
REALINTHANDLER(11,0x73)
REALINTHANDLER(12,0x74)
REALINTHANDLER(13,0x75)
REALINTHANDLER(14,0x76)
REALINTHANDLER(15,0x77)

	.text
	P2ALIGN(TEXT_ALIGN)
allintrs:
	pusha				/* save the general registers */
	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	/* See if we came from v86 mode.  */
	testl	$EFL_VM,TR_EFLAGS(%esp)
	jnz	intr_from_v86

intr_from_pmode:

	/* Load the kernel segment registers.  */
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* Patch the int instruction below.  */
	movl	TR_ERR(%esp),%eax
	movb	%al,2f+1

	/* Switch to v86 mode and handle the interrupt.  */
	ljmp	$KERNEL_16_CS,$1f
	.code16
1:	call	*EXT(i16_switch_to_real_mode)
2:	int	$0
	call	*EXT(i16_switch_to_pmode)
	ljmp	$KERNEL_CS,$return_from_trap
	.code32


/* Hardware interrupt from v86 mode.  Reflect it to DOS.  */
intr_from_v86:

	/* Load the linear/physical-address data segment,
	   for easy access to real-mode memory.  */
	movl	$LINEAR_DS,%eax
	movw	%ax,%ds

	/* Find the appropriate DOS interrupt vector,
	   stashed in the error slot by the interrupt stub.  */
	movl	TR_ERR(%esp),%esi

	/* Handle it just like a software interrupt.  */
	jmp	reflect_v86_intr


/*
 * Hardware interrupt handlers for reflecting interrupts to the protected mode program.
 */
#define DELAY							\
	jmp	9f						;\
9:	jmp	9f						;\
9:

#define ACK_FIRST(n)						\
	movb	$0x20,%al					;\
	outb	%al,$0x20

#define ACK_SECOND(n)						\
	movb	$0x20,%al					;\
	outb	%al,$0xa0					;\
	DELAY							;\
	outb	%al,$0x20

#define BLK_FIRST(n)						\
	inb	$0x21,%al					;\
	DELAY							;\
	orb	$1<<n,%al					;\
	outb	%al,$0x21					;\

#define BLK_SECOND(n)						\
	inb	$0xa1,%al					;\
	DELAY							;\
	orb	$1<<(n-8),%al					;\
	outb	%al,$0xa1					;\

#define UNBLK_FIRST(n)						\
	inb	$0x21,%al					;\
	DELAY							;\
	andb	$-1-(1<<n),%al					;\
	outb	%al,$0x21

#define UNBLK_SECOND(n)						\
	inb	$0xa1,%al					;\
	DELAY							;\
	andb	$-1-(1<<(n-8)),%al				;\
	outb	%al,$0xa1

#define	SIG_INTHANDLER(n,pic)					\
	.data							;\
	.long	1f						;\
	.data	2						;\
	.long	2f						;\
	.data	3						;\
	.long	3f						;\
	.text							;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* raw_sig_int_handler */				;\
1:	pushl	$0						;\
	pushl	$0x20+n						;\
	pusha							;\
	leal	EXT(sig_irq_nodes)+(SN_SIZE*n),%ebx		;\
	BLK_##pic(n)						;\
	DELAY							;\
	ACK_##pic(n)						;\
	jmp	sig_intrs					;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* dos_sig_int_handler */				;\
	.code16							;\
2:	pushl	%eax						;\
	pushl	%ebx						;\
	pushl	%ecx						;\
	leal	EXT(sig_irq_nodes)+(SN_SIZE*n),%ebx		;\
	BLK_##pic(n)						;\
	DELAY							;\
	ACK_##pic(n)						;\
	jmp	i16_dos_sig_intrs				;\
	.code32							;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* raw_sig_int_post_handler */				;\
3:	UNBLK_##pic(n)						;\
	ret

	.data
DATA(raw_sig_int_handlers)
	.data	2
DATA(dos_sig_int_handlers)
	.data	3
DATA(raw_sig_int_post_handlers)

SIG_INTHANDLER(0,FIRST)
SIG_INTHANDLER(1,FIRST)
SIG_INTHANDLER(2,FIRST)
SIG_INTHANDLER(3,FIRST)
SIG_INTHANDLER(4,FIRST)
SIG_INTHANDLER(5,FIRST)
SIG_INTHANDLER(6,FIRST)
SIG_INTHANDLER(7,FIRST)
SIG_INTHANDLER(8,SECOND)
SIG_INTHANDLER(9,SECOND)
SIG_INTHANDLER(10,SECOND)
SIG_INTHANDLER(11,SECOND)
SIG_INTHANDLER(12,SECOND)
SIG_INTHANDLER(13,SECOND)
SIG_INTHANDLER(14,SECOND)
SIG_INTHANDLER(15,SECOND)

	.text
	P2ALIGN(TEXT_ALIGN)
sig_intrs:
	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	/* Load the kernel segment registers.  */
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* Add the appropriate signode (ebx) to the pending queue.  */
	lea	EXT(sig_pending_list),%eax
	movl	SN_NEXT(%eax),%ecx
	movl	%ecx,SN_NEXT(%ebx)
	movl	%eax,SN_PREV(%ebx)
	movl	%ebx,SN_NEXT(%eax)
	movl	%ebx,SN_PREV(%ecx)

	/* Indicate that new signals need to be checked.  */
	movl	%esp,EXT(new_sigs)

	jmp	return_from_trap


	.code16

	/* If we're fielding the interrupt in DOS (v86) mode,
	   there's no chance we'll be able to dispatch the signal handler immediately,
	   so we don't need to save a full trap state frame like we do above.  */
	P2ALIGN(TEXT_ALIGN)
i16_dos_sig_intrs:

	/* Add the appropriate signode (ebx) to the pending queue.  */
	leal	EXT(sig_pending_list),%eax
	movl	%cs:SN_NEXT(%eax),%ecx
	movw	%cx,%cs:SN_NEXT(%ebx)
	movw	%ax,%cs:SN_PREV(%ebx)
	movw	%bx,%cs:SN_NEXT(%eax)
	movw	%bx,%cs:SN_PREV(%ecx)

	/* Indicate that new signals need to be checked.  */
	movw	%sp,%cs:EXT(new_sigs)

	/* Restore saved state and return.  */
	popl	%ecx
	popl	%ebx
	popl	%eax
	iretw


	.code32

#ifndef ENABLE_VCPI
#define vcpi_fastint_common -1
#endif

#define	FASTINTHANDLER(n,pic)					\
	.data							;\
	.long	1f						;\
	.data	2						;\
	.long	2f						;\
	.text							;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* raw_fastint_handler */				;\
1:	pushl	%eax						;\
	ACK_##pic(n)						;\
	leal	EXT(raw_fastint_nodes)+(8*n),%eax		;\
	jmp	raw_fastint_common				;\
	P2ALIGN(TEXT_ALIGN)					;\
	/* dos_fastint_handler */				;\
	.code16							;\
2:	pushl	%eax						;\
	pushl	%ecx						;\
	ACK_##pic(n)						;\
	leal	EXT(raw_fastint_nodes)+(8*n),%ecx		;\
	jmp	vcpi_fastint_common				;\
	.code32

	.data
DATA(raw_fastint_handlers)
	.data	2
DATA(vcpi_fastint_handlers)

FASTINTHANDLER(0,FIRST)
FASTINTHANDLER(1,FIRST)
FASTINTHANDLER(2,FIRST)
FASTINTHANDLER(3,FIRST)
FASTINTHANDLER(4,FIRST)
FASTINTHANDLER(5,FIRST)
FASTINTHANDLER(6,FIRST)
FASTINTHANDLER(7,FIRST)
FASTINTHANDLER(8,SECOND)
FASTINTHANDLER(9,SECOND)
FASTINTHANDLER(10,SECOND)
FASTINTHANDLER(11,SECOND)
FASTINTHANDLER(12,SECOND)
FASTINTHANDLER(13,SECOND)
FASTINTHANDLER(14,SECOND)
FASTINTHANDLER(15,SECOND)

	.text
	P2ALIGN(TEXT_ALIGN)
raw_fastint_common:
	pushl	%ecx
	pushl	%edx
	pushl	%ds
	pushl	%es

	/* Load the fastint ds/es registers
	   so we can push things onto the fastint stack.  */
	movl	$FASTINT_DS,%edx
	movw	%dx,%ds
	movw	%dx,%es

	/* Find the user's fastint handler parameters.
	   The handler ESP has already been adjusted 16 bytes downward
	   to make room for the stuff we're going to push onto it.  */
	movl	%ss:(%eax),%ecx		/* handler EIP */
	movl	%ss:4(%eax),%edx	/* handler ESP */

	/* Set up the long return stack frame for entry into the handler.  */
	movl	%ecx,(%edx)
	movl	$FASTINT_CS,4(%edx)

	/* The return pointer at offset 8 is permanently set
	   to point to the ljmp instruction in the program's crt0.  */

	/* Stash our current stack pointer so we'll be able to get back.  */
	movl	%esp,12(%edx)

	/* Switch to the handler stack.  */
	movl	%edx,%esp
	movl	$FASTINT_DS,%edx
	movw	%dx,%ss

	/* Drop into the handler.  */
	cld
	lret

ENTRY(raw_fastint_return)

	/* Get back to our kernel stack.  */
	movl	(%esp),%esp
	movl	$KERNEL_DS,%eax
	movw	%ax,%ss

	/* Restore the saved state and return from the interrupt.  */
	popl	%es
	popl	%ds
	popl	%edx
	popl	%ecx
	popl	%eax
	iret

#ifdef ENABLE_VCPI
	.code16

	P2ALIGN(TEXT_ALIGN)
vcpi_fastint_common:
	pushl	%edx
	pushl	%esi

	/* Now push the state frame we'll use to get back to v86 mode.  */
	movl	%esp,%eax
	pushl	%gs
	pushl	%fs
	pushl	%ds
	pushl	%es
	pushl	%ss
	pushl	%eax		/* esp */
	pushfl
	pushl	%cs
	pushl	$1f

	/* Find the stack top as a linear address (edx),
	   because that's what we'll need coming back out.  */
	xorl	%edx,%edx
	xorl	%eax,%eax
	movw	%ss,%dx
	movw	%sp,%ax
	shll	$4,%edx
	addl	%eax,%edx

	/* Jump into protected mode to handle the interrupt.
	   We'll wind up in vcpi_fastint_dispatcher, below.  */
	movw	$0xde0c,%ax
	movl	%cs:(vcpi_fastint_switch_data_la),%esi
	int	$0x67

	/* Just coming back out of protected mode.
	   Pop the saved state and return from the interrupt.  */
1:	popl	%esi
	popl	%edx
	popl	%ecx
	popl	%eax
	iretw

	.code32

ENTRY(vcpi_fastint_dispatcher)

	/* eax = scratch
	   ecx = kva of fastint node
	   edx = linear address of real-mode stack top
	   esi = scratch */

	/* Load the fastint ds/es registers
	   so we can push things onto the fastint stack.  */
	movl	$FASTINT_DS,%eax
	movw	%ax,%ds
	movw	%ax,%es
	movw	%ax,%ss

	/* Find the user's fastint handler parameters.
	   The handler ESP has already been adjusted 16 bytes downward
	   to make room for the stuff we're going to push onto it.  */
	movl	%cs:(%ecx),%eax		/* handler EIP */
	movl	%cs:4(%ecx),%esp	/* handler ESP */

	/* Set up the long return stack frame for entry into the handler.  */
	movl	%eax,(%esp)
	movl	$FASTINT_CS,4(%esp)

	/* Adjust the return pointer at offset 8
	   to point to the VCPI return bounce stub
	   instead of the normal protected-mode one.  */
	addl	$8,8(%esp)

	/* Stash our real-mode stack pointer so we'll be able to get back.  */
	movl	%edx,12(%esp)

	/* Drop into the handler.  */
	cld
	lret

ENTRY(vcpi_fastint_return)

	/* Re-adjust the return pointer
	   to point back to the normal bounce stub.  */
	subl	$8,-4(%esp)

	/* Switch back to the real-mode stack,
	   using LINEAR_DS and the linear address we saved.  */
	movl	(%esp),%esp
	movl	$LINEAR_DS,%eax
	movw	%ax,%ss

	/* Switch back to v86 mode.  */
	movw	%ax,%ds
	movl	$0xde0c,%eax
	lcall	%cs:EXT(vcpi_pmode_entry)

#endif /* ENABLE_VCPI */


/* System call in raw or vcpi mode.  */
ENTRY(raw_syscall_entry)

	/* The kernel expects to be uninterruptible.  */
	/*XXX*/
	cli

	/* Check for the special sigreturn syscall.  */
	cmpl	$-1,%eax
	jz	sigreturn

	/* Load the kernel data segments.  */
	movw	%ss,%cx
	movw	%cx,%ds
	movw	%cx,%es

	/* Clear the kernel's errno variable
	   so we'll know if an error occurred during the syscall.  */
	movl	$0,EXT(errno)

	/* Call the appropriate syscall routine,
	   with a pointer to the user-space arguments.  */
	movl	2*4(%esp),%ecx
	addl	EXT(prog_va),%ecx
	addl	$4,%ecx
	pushl	%ecx
	call	*EXT(syscall_table)(,%eax,4)
	addl	$4,%esp

	/* Load the final errno value to return to the caller.  */
	movl	EXT(errno),%edx

	/* Check for new signals.  */
	movl	EXT(new_sigs),%ecx
	orl	%edx,%ecx
	jnz	1f

	/* Reload the user data segments and return.  */
	movl	$USER_DS,%ecx
	movw	%cx,%ds
	movw	%cx,%es
	sti
	lret

/* Signals to handle - convert the syscall stack frame to a trap frame.  */
1:
	popl	%ecx		/* eip */
	popl	%edx		/* cs */
	pushfl
	orl	$EFL_IF,(%esp)	/* XXX */
	pushl	%edx
	pushl	%ecx
	pushl	$0		/* err */
	pushl	$-1		/* trapno */
	pushl	%eax		/* syscall return value */
	pushl	$0		/* ecx */
	pushl	$0		/* syscall errno value */
	pushl	%ebx
	pushl	$0		/* cr2 */
	pushl	%ebp
	pushl	%esi
	pushl	%edi
	pushl	$USER_DS
	pushl	$USER_DS
	pushl	%fs
	pushl	%gs

check_sigs_and_return:
#if 0 /* read the PIT */
	xorb	%al,%al
	outb	%al,$0x43
	outb	%al,$0x80
	inb	$0x40,%al
	xorl	%ecx,%ecx
	movb	%al,%cl
	outb	%al,$0x80
	inb	$0x40,%al
	movb	%al,%ch
	movl	ts1,%eax
	movl	%eax,%edx
	shll	$4,%eax
	subl	%edx,%eax
	shrl	$4,%eax
	addl	%ecx,%eax
	movl	%eax,ts1
#endif
	pushl	$return_from_trap
	jmp	EXT(check_sigs)

	/* Kernel esp points to the 16-byte i386 syscall frame;
	   user esp points just past sc_return_addr in the sigcontext.
	   The user segment registers are still loaded.  */
sigreturn:

	/* Turn the syscall frame into an interrupt frame.  */
	popl	%ecx			/* eip: throw away */
	popl	%edx			/* cs */
	popl	%eax			/* esp */
	addl	$SC_SIZE-4,%eax		/* pop the sigcontext */
	pushl	%eax			/* esp */
	pushl	-SC_SIZE+SC_EFL(%eax)
	pushl	%edx			/* cs */
	pushl	-SC_SIZE+SC_EIP(%eax)	/* eip: original eip before the signal */
	pushl	$0			/* err */
	pushl	$-1			/* trapno */
	pushl	-SC_SIZE+SC_EAX(%eax)
	pushl	-SC_SIZE+SC_ECX(%eax)
	pushl	-SC_SIZE+SC_EDX(%eax)
	pushl	%ebx			/* ebx */
	pushl	$0			/* cr2 */
	pushl	%ebp
	pushl	%esi
	pushl	%edi
	pushl	$USER_DS
	pushl	$USER_DS
	pushl	%fs
	pushl	%gs

	/* Grab the signode (edx) and the original signal mask (ecx).  */
	movl	-SC_SIZE+SC_SIGNODE(%eax),%edx
	movl	-SC_SIZE+SC_MASK(%eax),%ecx

	/* Load the kernel data segments.  */
	movw	%ss,%ax
	movw	%ax,%ds
	movw	%ax,%es

	/* Restore the original signal mask.  */
	movl	%ecx,EXT(sig_mask)

	/* Dispose of the signode in the appropriate way.  */
	pushl	%edx
	call	*SN_POST_HANDLER(%edx)
	addl	$4,%esp

	/* Check for more signals and then return.  */
	pushl	$return_from_trap
	jmp	EXT(check_sigs)


/*
 * Copy from user address space.
 * arg0:	user address
 * arg1:	kernel address
 * arg2:	byte count
 */
ENTRY(copyin)
	pushl	%esi
	pushl	%edi			/* save registers */

	movl	8+S_ARG0,%esi		/* get user start address */
	movl	8+S_ARG1,%edi		/* get kernel destination address */
	movl	8+S_ARG2,%edx		/* get count */

	movl	$copyin_fail,recovery

	movl	$USER_DS,%eax		/* use user data segment for accesses */
	mov	%ax,%ds

	/*cld*/				/* count up: default mode in all GCC code */
	movl	%edx,%ecx		/* move by longwords first */
	shrl	$2,%ecx
	rep
	movsl				/* move longwords */
	movl	%edx,%ecx		/* now move remaining bytes */
	andl	$3,%ecx
	rep
	movsb

	xorl	%eax,%eax		/* return 0 for success */

copyin_ret:
	mov	%ss,%di			/* restore DS to kernel segment */
	mov	%di,%ds

	movl	$0,recovery

	popl	%edi			/* restore registers */
	popl	%esi
	ret				/* and return */

copyin_fail:
	movl	$1,%eax			/* return 1 for failure */
	jmp	copyin_ret		/* pop frame and return */


/*
 * Copy to user address space.
 * arg0:	kernel address
 * arg1:	user address
 * arg2:	byte count
 */
ENTRY(copyout)
	pushl	%esi
	pushl	%edi			/* save registers */

	movl	8+S_ARG0,%esi		/* get kernel start address */
	movl	8+S_ARG1,%edi		/* get user start address */
	movl	8+S_ARG2,%edx		/* get count */

	movl	$copyout_fail,recovery

	movl	$USER_DS,%eax		/* use user data segment for accesses */
	mov	%ax,%es

	/*cld*/				/* count up: always this way in GCC code */
	movl	%edx,%ecx		/* move by longwords first */
	shrl	$2,%ecx
	rep
	movsl
	movl	%edx,%ecx		/* now move remaining bytes */
	andl	$3,%ecx
	rep
	movsb				/* move */

	xorl	%eax,%eax		/* return 0 for success */

copyout_ret:
	mov	%ss,%di			/* restore ES to kernel segment */
	mov	%di,%es

	movl	$0,recovery

	popl	%edi			/* restore registers */
	popl	%esi
	ret				/* and return */

copyout_fail:
	movl	$1,%eax			/* return 1 for failure */
	jmp	copyout_ret		/* pop frame and return */


