/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <a.out.h>
#include <errno.h>

#include <mach/exec/exec.h>
#include <mach/machine/proc_reg.h>

#include "dos_io.h"
#include "prog.h"
#include "dpmi.h"
#include "misc.h"
#include "debug.h"
#include "alloc.h"
#include "remote-gdb.h"

/* Environment-specific code must set these
   before any syscalls can get invoked by the program.  */
vm_offset_t prog_va;
const struct progvec *progvec;

struct prog_seg prog_seg[MAX_PROG_SEGS];
int prog_seg_count;

exec_info_t prog_info;

char *prog_filename = "a.out";

char *dos_buf;

static int fd;

static int e_read(void *handle, vm_offset_t file_ofs, void *buf, vm_size_t size, vm_size_t *actp)
{
	if (lseek(fd, file_ofs, SEEK_SET) < 0)
		return errno;
	if ((*actp = read(fd, buf, size)) < 0)
		return errno;
	return 0;
}

static int e_read_exec(void *handle, vm_offset_t file_ofs, vm_size_t file_size,
                       vm_offset_t mem_addr, vm_size_t mem_size, vm_prot_t mem_prot)
{
	if (prog_seg_count >= MAX_PROG_SEGS)
		return 12345678; /*XXX*/

	prog_seg[prog_seg_count].file_start = file_ofs;
	prog_seg[prog_seg_count].file_size = file_size;
	prog_seg[prog_seg_count].mem_start = mem_addr;
	prog_seg[prog_seg_count].mem_size = mem_size;
	prog_seg[prog_seg_count].mem_prot = mem_prot;

	prog_seg_count++;

	return 0;
}

void prog_start()
{
	/* Allocate our DOS bounce buffer so DOS I/O will work.  */
	dos_buf = alloc_dos(DOS_BUF_SIZE, 0, 0);
	if (!dos_buf)
		die("not enough low memory for DOS bounce buffer");

	save_video_mode();
	efs_init();
	vfs_init();
	sig_init();
	rgdb_init();

	if ((fd = open(prog_filename, O_RDONLY)) < 0)
		die("can't open program file `%s' (error 0x%x)",
		    prog_filename, errno);
	assert(fd > 0);
	if (errno = exec_load(e_read, e_read_exec, 0, &prog_info))
		die("can't execute program `%s' (error 0x%x)",
		    prog_filename, errno);

	/* This loads and starts running the program.  */
	progvec->run();
}

void prog_read(unsigned offset, void *start, unsigned size)
{
	int err;
	vm_offset_t newpos;
	vm_size_t actual;

	if (lseek(fd, offset, SEEK_SET) < 0)
		die("error seeking in program file `%s'", prog_filename);
	if (read(fd, start, size) != size)
		die("error reading program file `%s'", prog_filename);
}

static vm_offset_t push_str_array(vm_offset_t prog_esp, char **arr,
				  vm_offset_t *out_pva, int *out_count)
{
	int count, i;
	vm_offset_t arr_pva;
	vm_offset_t *arr_kva;

	/* Count the number of strings in the array.  */
	for (count = 0; arr[count]; count++);

	/* Allocate the array of pointers.  */
	prog_esp -= sizeof(char*) * (count+1);
	arr_pva = prog_esp;
	arr_kva = (vm_offset_t*)progva_to_va(arr_pva);

	/* Allocate and copy each string.  */
	for (i = 0; i < count; i++)
	{
		prog_esp -= strlen(arr[i])+1;
		prog_esp &= ~3;
		arr_kva[i] = prog_esp;
		strcpy((char*)progva_to_va(prog_esp), arr[i]);
	}

	/* Add the terminating null pointer.  */
	arr_kva[count] = 0;

	*out_pva = arr_pva;
	if (out_count) *out_count = count;
	return prog_esp;
}

vm_offset_t prog_push_args_env(vm_offset_t prog_esp)
{
	extern char **argv, **environ;
	vm_offset_t argv_pva, envp_pva;
	int argc;
	vm_offset_t *args_kva;

	prog_esp = push_str_array(prog_esp, argv, &argv_pva, &argc);
	prog_esp = push_str_array(prog_esp, environ, &envp_pva, 0);

	prog_esp -= 12;
	args_kva = (vm_offset_t*)progva_to_va(prog_esp);
	args_kva[0] = argc;
	args_kva[1] = argv_pva;
	args_kva[2] = envp_pva;

	return prog_esp;
}

