/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#ifndef _PROG_H
#define _PROG_H

#include <mach/machine/vm_param.h>
#include <mach/vm_prot.h>
#include <mach/exec/exec.h>

/* Size of the program's stack.  */
#define PROG_STACK_SIZE (1024*1024)

/* Address in our virtual address space
   of address 0 in the subject program's address space.  */
extern vm_offset_t prog_va;

#define progva_to_va(progva)	((typeof(progva))((int)(progva) + prog_va))
#define va_to_progva(va)	((typeof(va))((int)(va) - prog_va))

#define progva_to_la(progva)	((typeof(progva))((int)(progva) + prog_va - phys_mem_va))
#define la_to_progva(la)	((typeof(la))((int)(la) + phys_mem_va - prog_va))

/* Subject programs's virtual address map.
   These addresses are guaranteed to be in ascending order.
   The text and data segments are guaranteed not to overlap each other
   on the same page.  */
struct prog_seg
{
	vm_offset_t mem_start;
	vm_offset_t file_start;
	vm_size_t mem_size;
	vm_size_t file_size;
	vm_prot_t mem_prot;
};

#define MAX_PROG_SEGS 2

extern struct prog_seg prog_seg[MAX_PROG_SEGS];
extern int prog_seg_count;

/* Where the program should start executing.  */
extern exec_info_t prog_info;

/* Environment-specific code must set these.  */
struct signode;
struct progvec
{
	void (*run)(void);
	int (*sbrk)(int mem_needed, vm_offset_t *out_prog_va);
	int (*mmap_phys)(vm_offset_t pa, vm_size_t size, vm_offset_t *out_va);
	int (*mlock)(vm_offset_t prog_va, vm_size_t len);
	int (*munlock)(vm_offset_t prog_va, vm_size_t len);
	int (*mlockall)(int flags);
	int (*munlockall)(void);
	int (*irq_catch)(int irq, struct signode *n);
	void (*irq_free)(int irq);
	int (*irq_catch_fast)(int irq, vm_offset_t eip, vm_offset_t sp);
	void (*irq_free_fast)(int irq);
};
extern const struct progvec *progvec;

/* Filename of the program to load and run.  */
extern char *prog_filename;

void prog_start();
void prog_read(unsigned offset, void *start, unsigned size);
vm_offset_t prog_push_args_env(vm_offset_t prog_esp);

#endif /* _PROG_H */
