/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <mach/machine/proc_reg.h>
#include <mach/machine/pio.h>
#include <mach/machine/eflags.h>
#include <mach/machine/far_ptr.h>
#include <mach/machine/paging.h>
#include <mach/machine/vcpi.h>
#include <sys/signal.h>
#include <sys/mman.h>
#include <errno.h>

#include "vm_param.h"
#include "raw.h"
#include "ldt.h"
#include "prog.h"
#include "debug.h"
#include "config.h"
#include "remote-gdb.h"

/*XXX*/
#define INTEL_PTE_WIRED		0x00000200

/* This is always page-aligned.  */
static vm_offset_t prog_heap;

/* This isn't necessarily page-aligned;
   sbrk allows the break pointer to be incremented by partial pages.
   (That's how BSD malloc keeps blocks naturally aligned.)  */
static vm_offset_t prog_brk;

/* True if the program wants all allocated memory to be wired.  */
static int mlock_future;

static int raw_prog_mlock(vm_offset_t pva, vm_size_t len);

static void raw_prog_run()
{
	int i;
	extern unsigned short environ_seg, psp_seg;
	vm_offset_t prog_esp;

	/* Find the start of the stack/heap -
	   just above the highest program segment.  */
	prog_heap = 0;
	for (i = 0; i < MAX_PROG_SEGS; i++)
	{
		if ((prog_seg[i].mem_size > 0)
		    && (prog_heap < prog_seg[i].mem_start+prog_seg[i].mem_size))
		{
			prog_heap = prog_seg[i].mem_start + prog_seg[i].mem_size;
		}
	}
	prog_heap = round_page(prog_heap);

	/* Bump the heap segment start up a page
	   to leave a red zone below the stack.  */
	prog_heap += PAGE_SIZE;

	/* Allocate the stack at the base of the heap.  */
	prog_brk = prog_heap + PROG_STACK_SIZE;

	/* Set up prog_va appropriately.  */
	prog_va = phys_mem_va + PROG_MIN_LIN;

	/* Push the program's arguments and environment onto its stack.  */
	prog_esp = prog_push_args_env(prog_brk);

	/* Enable interrupts during program execution.  */
	sti();

	/* Drop into the program.  */
	asm volatile("
		pushl	%3
		pushl	%2
		pushl	%4
		pushl	%1
		pushl	%0
		movw	%3,%%ds
		movw	%3,%%es
		iret
	" :
	  : "rm" (prog_info.entry), "i" (USER_CS),
	    "rm" (prog_esp), "rm" ((unsigned short)USER_DS),
	    "r" (get_eflags() | (rgdb_enable ? EFL_TF : 0)));

	die("raw_prog_run");
}

static int raw_prog_sbrk(int mem_needed, vm_offset_t *out_prog_va)
{
	vm_offset_t va = prog_brk;
	int err;

	/* Bump the break pointer as requested.
	   (It's OK if the break pointer isn't page aligned;
	   it's OK if it isn't even bumped onto a different page.)  */
	prog_brk += mem_needed;

	/* If the program requested that all allocations be locked, do it.  */
	if (mlock_future)
	{
		if (err = raw_prog_mlock(va, mem_needed))
		{
			/* Roll back the brk pointer and return the error.  */
			prog_brk = va;
			return err;
		}
	}

	*out_prog_va = va;
	return 0;
}

static int raw_prog_mmap_phys(vm_offset_t pa, vm_size_t size, vm_offset_t *out_va)
{
	vm_offset_t page_pa = trunc_page(pa);
	vm_size_t page_size = round_page((pa - page_pa) + size);

	prog_brk = round_page(prog_brk);
	*out_va = prog_brk + (pa - page_pa);

	while (page_size)
	{
		int err;
		if (err = pmap_map_page(PROG_MIN_LIN+prog_brk,
			page_pa | INTEL_PTE_VALID | INTEL_PTE_USER
			| INTEL_PTE_WRITE))
			return err;
		page_pa += PAGE_SIZE;
		prog_brk += PAGE_SIZE;
		page_size -= PAGE_SIZE;
	}

	return 0;
}

static int raw_prog_mlock(vm_offset_t pva, vm_size_t len)
{
	vm_offset_t page_la;

	for (page_la = trunc_page(progva_to_la(pva));
	     page_la < round_page(progva_to_la(pva+len));
	     page_la += PAGE_SIZE)
	{
		pt_entry_t entry = pmap_get_pte(page_la);
		int err;

		if (entry & INTEL_PTE_VALID)
			err = pmap_map_page(page_la, entry | INTEL_PTE_WIRED);
		else
			err = page_fault(page_la, INTEL_PTE_WIRED);
		if (err) return err;
	}

	return 0;
}

static int raw_prog_munlock(vm_offset_t pva, vm_size_t len)
{
	vm_offset_t page_la;
	int err;

	for (page_la = trunc_page(progva_to_la(pva));
	     page_la < round_page(progva_to_la(pva+len));
	     page_la += PAGE_SIZE)
	{
		pt_entry_t entry = pmap_get_pte(page_la);
		if (entry & INTEL_PTE_VALID)
		{
			printf("unlocking la %08x\n", page_la);
			err = pmap_map_page(page_la, entry & ~INTEL_PTE_WIRED);
			if (err) return err;
		}
	}

	return 0;
}

static int raw_prog_mlockall(int flags)
{
	if (flags & MCL_FUTURE)
		mlock_future = 1;

	if (flags & MCL_CURRENT)
		panic("raw_prog_mlockall");

	return 0;
}

static int raw_prog_munlockall(void)
{
	panic("raw_prog_munlockall");
}

#include "idt.h"
#include "cpu.h"
#include "signal.h"
#include "real.h"

#ifdef ENABLE_VCPI
extern void v86_real_int(int intnum, struct real_call_data *rcd);
#define IN_VCPI_MODE() (real_int != v86_real_int)
#else
#define IN_VCPI_MODE() 0
#endif

extern vm_offset_t raw_sig_int_handlers[16];
extern vm_offset_t dos_sig_int_handlers[16];
extern void (*raw_sig_int_post_handlers[16])(struct signode *n);
extern vm_offset_t real_int_handlers[16];

static struct far_pointer_16 orig_dos_vecs[16];

static int raw_prog_irq_catch(int irq, struct signode *n)
{
	struct far_pointer_16 new_dos_vec;
	int intvec = (irq < 8) ? (irq_master_base + irq)
			       : (irq_slave_base + irq - 8);

	assert(irq >= 0); assert(irq < 16);
	assert(!(get_eflags() & EFL_IF));

	/* Set the protected-mode interrupt handler.  */
	fill_irq_gate(&cpu[0], irq, raw_sig_int_handlers[irq],
		      KERNEL_CS, ACC_PL_K);

	/* If we're in VCPI mode, hook the DOS interrupt vector as well.
	   This isn't necessary in raw mode
	   where we control the v86 box that DOS runs in;
	   then interrupts in v86 mode will still go
	   to our pmode interrupt handler.  */
	if (IN_VCPI_MODE())
	{
		new_dos_vec.seg = real_cs;
		new_dos_vec.ofs = dos_sig_int_handlers[irq];
		dos_get_int_vec(intvec, &orig_dos_vecs[irq]);
		dos_set_int_vec(intvec, &new_dos_vec);
	}

	/* This routine gets called after the user signal hander returns;
	   it reenables the hardware interrupt.  */
	n->post_handler = raw_sig_int_post_handlers[irq];

	return 0;
}

static void raw_prog_irq_free(int irq)
{
	int intvec = (irq < 8) ? (irq_master_base + irq)
			       : (irq_slave_base + irq - 8);

	assert(irq >= 0); assert(irq < 16);
	assert(!(get_eflags() & EFL_IF));

	/* Reset the protected-mode interrupt handler
	   to the normal reflection routine.  */
	fill_irq_gate(&cpu[0], irq, real_int_handlers[irq],
		      KERNEL_CS, ACC_PL_K);

	/* Reset the DOS interrupt vector.  */
	if (IN_VCPI_MODE())
		dos_set_int_vec(intvec, &orig_dos_vecs[irq]);
}

extern vm_offset_t raw_fastint_handlers[16];
extern void raw_fastint_return();

struct
{
	vm_offset_t eip;
	vm_offset_t esp;
} raw_fastint_nodes[16];

static struct vcpi_switch_data vcpi_fastint_switch_data;
vm_offset_t vcpi_fastint_switch_data_la;

static int raw_prog_irq_catch_fast(int irq, vm_offset_t eip, vm_offset_t esp)
{
	struct far_pointer_16 new_dos_vec;
	int intvec = (irq < 8) ? (irq_master_base + irq)
			       : (irq_slave_base + irq - 8);
	struct far_pointer_32 *bounce_code;

	assert(irq >= 0); assert(irq < 16);
	assert(!(get_eflags() & EFL_IF));

	/* Stash the handler eip and esp.
	   The fastint glue code in locore.S expects esp
	   to be adjusted by 16 bytes for the initial stack frame.  */
	raw_fastint_nodes[irq].eip = eip;
	raw_fastint_nodes[irq].esp = esp - 16;

	/* Fill in the return address in the stack frame,
	   which never changes from interrupt to interrupt.  */
	*((long*)progva_to_va(esp - 8)) = prog_info.entry - 32;

	/* Fill in the magic bounce code in the program's crt0.  */
	assert(*((unsigned char*)progva_to_va(prog_info.entry - 32)) == 0xea);
	bounce_code = (struct far_pointer_32*)
		progva_to_va(prog_info.entry - 32 + 1);
	bounce_code->ofs = (vm_offset_t)raw_fastint_return;
	bounce_code->seg = KERNEL_CS;

	/* Set the protected-mode interrupt handler.  */
	fill_irq_gate(&cpu[0], irq, raw_fastint_handlers[irq],
		      KERNEL_CS, ACC_PL_K);

	/* If we're in VCPI mode, hook the DOS interrupt vector as well.
	   This isn't necessary in raw mode
	   where we control the v86 box that DOS runs in;
	   then interrupts in v86 mode will still go
	   to our pmode interrupt handler.  */
	if (IN_VCPI_MODE())
	{
		extern vm_offset_t vcpi_fastint_handlers[16];
		extern struct vcpi_switch_data vcpi_switch_data;
		extern void vcpi_fastint_dispatcher();
		extern void vcpi_fastint_return();

		/* Fill in the VCPI bounce code in the program's crt0.  */
		assert(*((unsigned char*)
			progva_to_va(prog_info.entry - 24)) == 0xea);
		bounce_code = (struct far_pointer_32*)
			progva_to_va(prog_info.entry - 24 + 1);
		bounce_code->ofs = (vm_offset_t)vcpi_fastint_return;
		bounce_code->seg = KERNEL_CS;

		/* Set up the mode switching data structure.  */
		vcpi_fastint_switch_data = vcpi_switch_data;
		vcpi_fastint_switch_data.entry_eip =
			(vm_offset_t)vcpi_fastint_dispatcher;
		vcpi_fastint_switch_data.entry_cs = KERNEL_CS;
		vcpi_fastint_switch_data_la =
			(vm_offset_t)kvtolin(&vcpi_fastint_switch_data);

		/* Hook the appropriate DOS interrupt vector.  */
		new_dos_vec.seg = real_cs;
		new_dos_vec.ofs = vcpi_fastint_handlers[irq];
		dos_get_int_vec(intvec, &orig_dos_vecs[irq]);
		dos_set_int_vec(intvec, &new_dos_vec);
	}

	return 0;
}

const struct progvec raw_progvec =
{
	raw_prog_run,
	raw_prog_sbrk,
	raw_prog_mmap_phys,
	raw_prog_mlock,
	raw_prog_munlock,
	raw_prog_mlockall,
	raw_prog_munlockall,
	raw_prog_irq_catch,
	raw_prog_irq_free,
	raw_prog_irq_catch_fast,
	raw_prog_irq_free,
};

int page_fault(vm_offset_t linear_address, int pteflags)
{
	vm_offset_t va = linear_address - PROG_MIN_LIN;
	vm_offset_t page_va = trunc_page(va);
	vm_offset_t page_pa;
	struct prog_seg *s;
	int seg;
	int err;

	pteflags |= INTEL_PTE_VALID | INTEL_PTE_USER;

	if (linear_address < PROG_MIN_LIN)
		return EFAULT;

	/* If we're accessing the heap, just provide a zero-filled page.  */
	if ((va >= prog_heap) && (va < round_page(prog_brk)))
	{
		page_pa = page_alloc();
		if (!page_pa)
			return ENOMEM;
		bzero(phystokv(page_pa), PAGE_SIZE);
		if (err = pmap_map_page(PROG_MIN_LIN+page_va,
				        page_pa | pteflags | INTEL_PTE_WRITE))
			return err; /* XXX we lose the page */
		return 0;
	}

	/* Find the segment that provides this page.  */
	assert(prog_seg_count <= MAX_PROG_SEGS);
	for (seg = 0; ; seg++)
	{
		if (seg >= prog_seg_count)
			return EFAULT;
		s = &prog_seg[seg];
		if ((va >= s->mem_start) && (va < s->mem_start+s->mem_size))
			break;
	}

	/* Allocate and initialize the page.  */
	page_pa = page_alloc();
	if (!page_pa)
		return ENOMEM;
	if ((page_va >= s->mem_start + s->file_size)
	    || (s->file_size == 0))
	{
		/* Just zero fill the whole page.  */
		bzero(phystokv(page_pa), PAGE_SIZE);
	}
	else
	{
		/* Have to load something into the page.  */
		vm_offset_t ofs = page_va - s->mem_start + s->file_start;
		vm_size_t amt = PAGE_SIZE;

		/* See if we need to zero fill the end of the page.  */
		if (page_va + PAGE_SIZE > s->mem_start + s->file_size)
		{
			amt = (s->mem_start + s->file_size) - page_va;
			bzero(phystokv(page_pa + amt), PAGE_SIZE - amt);
		}
		assert(amt > 0); assert (amt <= PAGE_SIZE);

		/* Load the contents of the page.  */
		assert(ofs >= 0);
		prog_read(ofs, (void*)phystokv(page_pa), amt);
	}
	if (s->mem_prot & VM_PROT_WRITE)
		pteflags |= INTEL_PTE_WRITE;

	/* Map the page into the program's address space.  */
	if (err = pmap_map_page(PROG_MIN_LIN+page_va, page_pa | pteflags))
		return err; /* XXX we lose the page */

	return 0;
}

