/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <mach/machine/vm_param.h>

#include "real.h"
#include "dpmi.h"
#include "debug.h"

struct real_call_data real_call_data; /*XXX*/

void real_ptr_from_va(const void *ptr, unsigned short *seg, unsigned short *ofs)
{
	extern vm_offset_t phys_mem_va;
	extern unsigned short real_cs;

	/* We need a real-mode far pointer to the data.  */
	if ((vm_offset_t)ptr < 0x10000)
	{
		/* Pointer is into our program data or bss
		   (hopefully not text :-) ).
		   Use real_cs as the segment,
		   and the supplied pointer as the offset.  */

		extern char etext[];

		assert((vm_offset_t)ptr >= (vm_offset_t)etext);

		*seg = real_cs;
		*ofs = (unsigned short)(vm_offset_t)ptr;
	}
	else
	{
		/* Pointer had better point into mapped physical memory.  */

		vm_offset_t phys = (vm_offset_t)ptr - phys_mem_va;

		assert(phys < 0x100000);

		*seg = phys >> 4;
		*ofs = phys & 0xf;
	}
}

void real_set_ds_dx(const void *ptr)
{
	real_ptr_from_va(ptr,
			 &real_call_data.ds,
			 (unsigned short*)&real_call_data.edx);
}

void real_set_es_di(const void *ptr)
{
	real_ptr_from_va(ptr,
			 &real_call_data.es,
			 (unsigned short*)&real_call_data.edi);
}

void real_call_data_dump(void)
{
	printf("contents of real_call_data:\n");
	printf("eax %08x  ebx %08x  ecx %08x  edx %08x\n",
		real_call_data.eax, real_call_data.ebx,
		real_call_data.ecx, real_call_data.edx);
	printf("esi %08x  edi %08x  ebp %08x\n",
		real_call_data.esi, real_call_data.edi,
		real_call_data.ebp);
	printf("sp %04x  ss %04x  flags %04x\n",
		real_call_data.ss, real_call_data.sp,
		real_call_data.flags);
}

