/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 *	Loosely based on i386-stub.c from GDB 4.14
 */

#include <mach/machine/vm_param.h>
#include <mach/machine/eflags.h>
#include <mach/machine/pio.h>

#include "prog.h"
#include "trap.h"
#include "signal.h"
#include "remote-gdb.h"

#ifdef ENABLE_REMOTE_GDB

/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
#define BUFMAX 400

/* This structure represents the register state frame as GDB wants it.  */
struct gdb_state
{
	unsigned	eax;
	unsigned	ecx;
	unsigned	edx;
	unsigned	ebx;
	unsigned	esp;
	unsigned	ebp;
	unsigned	esi;
	unsigned	edi;
	unsigned	eip;
	unsigned	eflags;
	unsigned	cs;
	unsigned	ss;
	unsigned	ds;
	unsigned	es;
	unsigned	fs;
	unsigned	gs;
};

#define ser_io_base	0x3f8	/* COM1 */
#define ser_irq		4

int rgdb_enable = 1;

static const char hexchars[]="0123456789abcdef";

static int hex(char ch)
{
	if ((ch >= '0') && (ch <= '9')) return (ch-'0');
	if ((ch >= 'a') && (ch <= 'f')) return (ch-'a'+10);
	if ((ch >= 'A') && (ch <= 'F')) return (ch-'A'+10);
	return (-1);
}

static int getDebugChar()
{
	/* Wait for a character to arrive.  */
	while (!(inb(ser_io_base + 5) & 0x01));

	/* Grab and return it.  */
	return inb(ser_io_base + 0);
}

static int putDebugChar(int ch)
{
	/* Wait for the transmit buffer to become available.  */
	while (!(inb(ser_io_base + 5) & 0x20));

	outb(ser_io_base + 0, ch);
}

/* Scan for the sequence $<data>#<checksum> */
static void getpacket(char *buffer)
{
	unsigned char checksum;
	unsigned char xmitcsum;
	int  i;
	int  count;
	char ch;

	do {
		/* Wait around for the start character;
		   ignore all other characters.  */
		while ((ch = (getDebugChar() & 0x7f)) != '$');
		checksum = 0;
		xmitcsum = -1;

		count = 0;

		/* Now, read until a # or end of buffer is found.  */
		while (count < BUFMAX) {
			ch = getDebugChar() & 0x7f;
			if (ch == '#') break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#')
		{
			xmitcsum = hex(getDebugChar() & 0x7f) << 4;
			xmitcsum += hex(getDebugChar() & 0x7f);
			if (checksum != xmitcsum)
				putDebugChar('-');  /* failed checksum */
			else
			{
				putDebugChar('+');  /* successful transfer */

				/* If a sequence char is present,
				   reply the sequence ID.  */
				if (buffer[2] == ':')
				{
					putDebugChar( buffer[0] );
					putDebugChar( buffer[1] );

					/* remove sequence chars from buffer */
					count = strlen(buffer);
					for (i=3; i <= count; i++)
						buffer[i-3] = buffer[i];
				}
			}
		}
	} while (checksum != xmitcsum);
}

/* send the packet in buffer.  */
static void putpacket(char *buffer)
{
	unsigned char checksum;
	int  count;
	char ch;

	/*  $<packet info>#<checksum>. */
	do {
		putDebugChar('$');
		checksum = 0;
		count    = 0;

		while (ch=buffer[count]) {
			if (! putDebugChar(ch)) return;
			checksum += ch;
			count += 1;
		}

		putDebugChar('#');
		putDebugChar(hexchars[checksum >> 4]);
		putDebugChar(hexchars[checksum % 16]);

	} while ((getDebugChar() & 0x7f) != '+');
}

static void bin2hex(unsigned char *in, char *out, int len)
{
	while (len--)
	{
		*out++ = hexchars[*in >> 4];
		*out++ = hexchars[*in++ % 16];
	}
	*out = 0;
}

static int hex2bin(char *in, unsigned char *out, int len)
{
	while (len--)
	{
		int ch = (hex(in[0]) << 4) | hex(in[1]);
		if (ch < 0)
			return -1;
		*out++ = ch;
		in += 2;
	}
	return 0;
}

void rgdb_init(void)
{
	struct sigevent ev;

	if (!rgdb_enable)
		return;

	/* Initialize the serial port.  */
	outb(ser_io_base + 3, 0x83);	/* DLAB = 1, frame = 8N1 */
	outb(ser_io_base + 0, 12);	/* 9600 baud */
	outb(ser_io_base + 1, 0);
	outb(ser_io_base + 3, 0x03);	/* DLAB = 0, frame = 8N1 */
	outb(ser_io_base + 1, 0x00);	/* no interrupts enabled */
	outb(ser_io_base + 4, 0x0b);	/* OUT2, RTS, and DTR enabled */

	/* Clear all serial interrupts.  */
	inb(ser_io_base + 6);	/* ID 0: read RS-232 status register */
	inb(ser_io_base + 2);	/* ID 1: read interrupt identification reg */
	inb(ser_io_base + 0);	/* ID 2: read receive buffer register */
	inb(ser_io_base + 5);	/* ID 3: read serialization status reg */

	/* Now enable the received data interrupt only */
	outb(ser_io_base + 1, 0x01);	/* interrupt on received data */

	/* Grab the appropriate serial interrupt line
	   so that the remote debugger can interrupt us during execution.  */
	ev.sigev_notify = SIGEV_SIGNAL;
	ev.sigev_signo = SIGINT;
	if (irq_request(ser_irq, 0, &ev))
		panic("rgdb_init: IRQ%d not available for remote debugging",
			ser_irq);
}

int rgdb_signal(int *inout_signo, struct trap_state *ts)
{
	int signo = *inout_signo;
	char inbuf[BUFMAX+1];
	char outbuf[BUFMAX+1];

	if (!rgdb_enable)
		return 0;

	/* If we were interrupted by the remote debugger,
	   then clear the serial interrupt.  */
	if (signo == SIGINT)
	{
		int iir = inb(ser_io_base + 2);
		if (iir & 0x01)
		{
			/* Spurious interrupt - just consume it and return.  */
			return 1;
		}
		if ((iir & 0x06) != 0x04)
			panic("unexpected serial interrupt iir=%d", iir);

		/* The interrupt will be cleared automatically
		   when we start polling for incoming packets.  */
	}

	/* Reply to the host that a signal has occurred.  */
	sprintf(outbuf, "S%02x", signo);
	putpacket(outbuf);

	while (1)
	{
		/* Return a blank response to messages we don't understand.  */
		outbuf[0] = 0;

		getpacket(inbuf);
		switch (inbuf[0])
		{
			case '?':
			{
				sprintf(outbuf, "S%02x", signo);
				break;
			}
			case 'g': /* return the value of the CPU registers */
			{
				struct gdb_state r;

				/* Convert the trap state to GDB's format.  */
				r.eax = ts->eax;
				r.ecx = ts->ecx;
				r.edx = ts->edx;
				r.ebx = ts->ebx;
				r.esp = ts->esp;
				r.ebp = ts->ebp;
				r.esi = ts->esi;
				r.edi = ts->edi;
				r.eip = ts->eip;
				r.eflags = ts->eflags;
				r.cs = ts->cs;
				r.ss = ts->ss;
				r.ds = ts->ds;
				r.es = ts->es;
				r.fs = ts->fs;
				r.gs = ts->gs;

				/* Send it over like a memory dump.  */
				bin2hex((char*)&r, outbuf, sizeof(r));

				break;
			}
			case 'G' : /* set the CPU registers - return OK */
			{
				struct gdb_state r;

				/* Unpack the new register state.  */
				if (hex2bin(&inbuf[1], (char*)&r, sizeof(r)))
				{
					strcpy(outbuf, "E04");
					break;
				}

				/* Stuff it into our trap_state.  */
				ts->eax = r.eax;
				ts->ecx = r.ecx;
				ts->edx = r.edx;
				ts->ebx = r.ebx;
				ts->esp = r.esp;
				ts->ebp = r.ebp;
				ts->esi = r.esi;
				ts->edi = r.edi;
				ts->eip = r.eip;
				ts->eflags = r.eflags;
				ts->cs = r.cs;
				ts->ss = r.ss;
				ts->ds = r.ds;
				ts->es = r.es;
				ts->fs = r.fs;
				ts->gs = r.gs;

				strcpy(outbuf, "OK");
				break;
			}

			/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
			case 'm' :
			{
				vm_offset_t addr, len;
				char *ptr, *ptr2;
				int i;

				/* Read the start address and length.
				   If invalid, send back an error code.  */
				strcpy(outbuf, "E01");
				addr = strtoul(&inbuf[1], &ptr, 16);
				if ((ptr == &inbuf[1]) || (*ptr != ','))
					break;
				ptr++;
				len = strtoul(ptr, &ptr2, 16);
				if ((ptr2 == ptr) || (len*2 > BUFMAX))
					break;

				/* Copy the data into a kernel buffer.
				   If a fault occurs, return an error code.  */
				if (copyin(addr, inbuf, len))
				{
					strcpy(outbuf, "E03");
					break;
				}

				/* Convert it to hex data in the outbuf.  */
				bin2hex(inbuf, outbuf, len);

				break;
			}

			/* MAA..AA,LLLL: Write LLLL bytes at AA.AA return OK */
			case 'M' :
			{
				vm_offset_t addr, len;
				char *ptr, *ptr2;
				int i;

				/* Read the start address and length.
				   If invalid, send back an error code.  */
				strcpy(outbuf, "E02");
				addr = strtoul(&inbuf[1], &ptr, 16);
				if ((ptr == &inbuf[1]) || (*ptr != ','))
					break;
				ptr++;
				len = strtoul(ptr, &ptr2, 16);
				if ((ptr2 == ptr) || (*ptr2 != ':'))
					break;
				ptr2++;
				if (ptr2 + len*2 > &inbuf[BUFMAX])
					break;

				/* Convert the hex input data to binary.  */
				if (hex2bin(ptr2, outbuf, len))
					break;

				/* Copy the data into its final place.
				   If a fault occurs, return an error code.  */
				if (copyout(outbuf, addr, len))
				{
					strcpy(outbuf, "E03");
					break;
				}

				strcpy(outbuf, "OK");
				break;
			}

			/* cAA..AA Continue at address AA..AA(optional) */
			/* sAA..AA Step one instruction from AA..AA(optional) */
			case 'c' :
			case 's' :
			{
				vm_offset_t new_eip;
				char *ptr;

				/* Try to read optional parameter;
				   leave EIP unchanged if none.  */
				new_eip = strtoul(&inbuf[1], &ptr, 16);
				if (ptr > &inbuf[1])
					ts->eip = new_eip;

				/* Set the trace flag appropriately.  */
				if (inbuf[0] == 's')
					ts->eflags |= EFL_TF;
				else
					ts->eflags &= ~EFL_TF;

				/* Return and "consume" the signal
				   that caused us to be called.  */
				*inout_signo = 0;
				return 1;
			}

			/* CssAA..AA Continue with signal ss */
			/* SssAA..AA Step one instruction with signal ss */
			case 'C' :
			case 'S' :
			{
				vm_offset_t new_eip;
				int nsig;
				char *ptr;

				/* Read the new signal number.  */
				nsig = hex(inbuf[1]) << 4 | hex(inbuf[2]);
				if ((nsig <= 0) || (nsig >= NSIG))
				{
					strcpy(outbuf, "E05");
					break;
				}

				/* Try to read optional parameter;
				   leave EIP unchanged if none.  */
				new_eip = strtoul(&inbuf[3], &ptr, 16);
				if (ptr > &inbuf[3])
					ts->eip = new_eip;

				/* Set the trace flag appropriately.  */
				if (inbuf[0] == 'S')
					ts->eflags |= EFL_TF;
				else
					ts->eflags &= ~EFL_TF;

				/* Return and "consume" the signal
				   that caused us to be called.  */
				*inout_signo = nsig;
				return 1;
			}

			case 'k': /* kill the program */
				panic("Program terminated by debugger");

		}

		/* reply to the request */
		putpacket(outbuf);
	}
}

void rgdb_shutdown(int exit_code)
{
	char outbuf[4];

	if (!rgdb_enable)
		return;

	/* Indicate that we have terminated.  */
	sprintf(outbuf, "W%02x", exit_code);
	putpacket(outbuf);
}

#endif /* ENABLE_REMOTE_GDB */
