/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1995 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef _MOSS_KERNEL_RAW_SIGNAL_H_
#define _MOSS_KERNEL_RAW_SIGNAL_H_

#include <sys/signal.h>
#include <mach/inline.h>

#include "config.h"

/* In-kernel structure representing a queued signal.  */
struct signode
{
	struct signode *next, *prev;
	int signo;
	sigset_t sigmask;
	union sigval value;
	void (*post_handler)(struct signode *n);
};

/* Context pushed onto the stack for a user signal handler.  */
struct sigcontext
{
	unsigned	sc_return_addr;
	unsigned	sc_signo;
	union sigval	sc_value;
	sigset_t	sc_mask;
	unsigned	sc_eip;
	unsigned	sc_efl;
	unsigned	sc_edx;
	unsigned	sc_ecx;
	unsigned	sc_eax;
	struct signode *sc_signode;
};

MACH_INLINE void sig_insert_after(struct signode *new_node, struct signode *old_node)
{
	new_node->next = old_node->next;
	new_node->prev = old_node;
	new_node->next->prev = new_node;
	new_node->prev->next = new_node;
}

MACH_INLINE void sig_remove(struct signode *node)
{
	node->next->prev = node->prev;
	node->prev->next = node->next;
}

/* List of pending signals.  */
extern struct signode sig_pending_list;

/* Current signal mask.  */
extern sigset_t sig_mask;

/* signodes for the hardware irqs.  */
extern struct signode sig_irq_nodes[16];

void sig_cleanup(void);
int irq_request(int irq, int pri, struct sigevent *ev);
int irq_free(int irq);

#endif _MOSS_KERNEL_RAW_SIGNAL_H_
