/* 
 * Copyright (c) 1996-1995 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *      Author: Bryan Ford, University of Utah CSL
 */

#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <mach/vm_param.h>

int
__sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp,
	 void *newp, size_t *newlen)
{
	if (namelen < 1)
		goto inval;
	switch (name[0])
	{
	case CTL_HW:
		if (namelen != 2)
			goto inval;
		switch (name[1])
		{
		case HW_PAGESIZE:
			if (newp) goto perm;
			if (!oldp) return 0;
			if (*oldlenp < sizeof(int)) goto nomem;
			*oldlenp = sizeof(int);

			*(int*)oldp = PAGE_SIZE;
			return 0;
		}
		break;
	}
	panic("__sysctl %d:%d %d %d", namelen, name[0], name[1], name[2]);

inval:
	errno = EINVAL;
	return -1;

perm:
	errno = EPERM;
	return -1;

nomem:
	errno = ENOMEM;
	return -1;
}

